; Copyright (c) 2022-2024 , Juergen Lorenz, ju (at) jugilo (dot) de
; All rights reserved.
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are
; met:
; 
; Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
; 
; Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following disclaimer in the
; documentation and/or other materials provided with the distribution.
; Neither the name of the author nor the names of its contributors may be
; used to endorse or promote products derived from this software without
; specific prior written permission. 
;   
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
; IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
; TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
; PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
; HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
; TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#|[
This module provides routines, to access and traverse trees,
i.e. nested pseudo-lists, as well as some list routines. They all
help writing pattern matching and procedural macros.

In particular, there is a replacement of car, cdr and consorts by one
operator, walk, of integer arguments, or, to be more precise, by +1 and
-1: (walk -1) is car, (walk 1) cdr, (walk) is identity.  These operaters
can (and will) be composed, so, for example, (walk -1 1 1) is
(o (walk -1) (walk 1) (walk 1)), in other words caddr.

The more arguments you give to walk, the deeper you can dig into a tree.
If you like, you can interprete the arguments of walk as a path for
traversing the tree. For compatibility with the c*r routines, you
interprete the arguments of walk from right to left. It might be more
convenient, to do it left to right; walk* does this.

Other routines, which walk a tree, are provided as well. In particular,
we provide a routine, tree->pathes, which maps a tree to a flat list of
pairs, containing the tree's items and the corresponding pathes to
access that items. This list can be used, to destructure similarly
shaped trees.

This way, we can use one tree as a pattern and take its pathes to
destructure another tree, i.e. another nested pseudolist. You gess it:
This will make pattern matching and writing procedural macros easy.

Some routines, which might be helpfull in macro-writing or
tree-analysing, are provided as well, for example map*, filter* and
friends.

]|#

(module tree-walkers (
  values->list
  list->values
  list-values
  always
  filter
  filter*
  filter-pairs
  at
  take
  drop
  split-at
  split-when
  take-while
  drop-while
  every?
  all?
  some?
  none?
  pseudo-list?
  pseudo-length
  pseudo-head
  pseudo-tail
  pseudo-list-of?
  indexes
  tree-of?
  iterate
  map*
  when-map*
  map**
  flatten*
  zip2
  unzip*
  walk
  walk*
  memp
  dups-remove
  and?
  sublist
  tree->pathes
  pathes
  same-shape?
  pseudo->accessors
  tree->accessors
  apply-accessors
  all-match?
  tree-walkers
  )

  (import scheme
          (only (chicken base)
                o error define-inline print case-lambda
                atom? complement assert
                receive list-of? identity)
          (only (chicken condition) condition-case)
          ;(only (chicken keyword) keyword?)
          ;simple-tests ;;;
          )

(define-inline (mappend fn lists)
  (apply append (map fn lists)))

;(define-inline (print-lines . lines)
;  (for-each print lines))

;;;; internal
;(define (make-list len val)
;  (let loop ((len len) (result '()))
;    (if (zero? len)
;      result
;      (loop (- len 1) (cons val result)))))

;(define (listify pair)
;  (cond
;    ((null? pair) '())
;    ((pair? pair)
;     (cons (car pair) (listify (cdr pair))))
;    (else
;      (list pair))))

#|[
(values->list vals)
--- macro ---
transforms values, vals, into a list
]|#
(define-syntax values->list
  (syntax-rules ()
    ((_ vals) (call-with-values (lambda () vals)
                                (lambda xs xs)))))

#|[
(list->values lst)
--- procedure ---
transforms a list, lst, into a values
]|#
(define (list->values lst)
  (apply values lst))

#|[
(list-values proc . args)
--- procedure ---
transforms the resulting values of calling proc
with args into a list.
Used in tests most of the time
]|#
(define (list-values proc . args)
  (receive vals (apply proc args)
    (apply list vals)))

#|[
(always xpr)
--- procedure ---
returns a procedure of arbitrary many arguments,
which always returns the value of xpr, whatever
it's arguments are given
]|#
(define (always xpr)
  (lambda args xpr))

#|[
(filter ok? lst)
--- procedure ---
splits a list into two sublists, one that pass the ok? test and one that
doesn't
]|#
(define (filter ok? xs)
  (let loop ((xs xs) (yes '()) (no '()))
    (cond
      ((pair? xs)
       (let ((x (car xs)) (xs (cdr xs)))
         (if (ok? x)
           (loop xs (cons x yes) no)
           (loop xs yes (cons x no)))))
      ((null? xs)
       (values (reverse yes) (reverse no)))
      (else
        (error 'filter "not a list" xs)))))

#|[
(filter* ok? tree)
--- procedure ---
filters a tree according to the predicate ok? respecting the tree
structure
]|#
(define (filter* ok? tree)
  (let loop ((tree tree))
    (cond
      ((null? tree) '())
      ((pair? (car tree))
       (cons (loop (car tree)) (loop (cdr tree))))
      ((ok? (car tree))
       (cons (car tree) (loop (cdr tree))))
      (else (loop (cdr tree))))))

#|[
(filter-pairs ok? pairs)
--- procedure ---
splits a list of pairs into two sublists, one that pass the ok? test
on the cars and one that doesn't
]|#
(define (filter-pairs ok? pairs)
  (filter (lambda (pair)
            (ok? (car pair)))
          pairs))

#|[
(at n)
(at n lst)
--- procedure ---
the second takes the nth  item of the list lst,
the first is a curried version of the second.
]|#
(define (at n . lists)
  (if (null? lists)
    (lambda (lst)
      (list-ref lst n))
    (list-ref (car lists) n)))

#|[
(take n)
(take n lst)
--- procedure ---
the second takes the first n items of the list lst,
the first is a curried version of the second.
]|#
(define (take n . lists)
  (if (null? lists)
    (lambda (lst)
      (split-at n lst))
    (split-at n (car lists))))

#|[
(drop n)
(drop n lst)
--- procedure ---
the second drops the first n items of the list lst,
the first is a curried version of the second.
]|#
(define (drop n . lists)
  (if (null? lists)
    (lambda (lst)
      (list-tail lst n))
    (list-tail (car lists) n)))

#|[
(split-at n pls)
--- procedure ---
splitting a pseudo-list in two sublists, head and tail, the sublist head
excluding the nth item and the pseudolist tail starting with nth item.
]|#
(define (split-at n pls)
  (let loop ((tail pls) (k 0) (head '()))
    (cond
      ((null? tail)
       (values (reverse head) tail))
      ((= k n)
       (values (reverse head) tail))
      (else
        (loop (cdr tail) (+ k 1) (cons (car tail) head))))))

#|[
(split-when ok? lst)
--- procedure ---
splitting a list in two sublists, head and tail, the sublist head before
the first item passing the ok? test, the sublist tail starting at
that item.
]|#
(define (split-when ok? lst)
  (let loop ((tail lst) (head '()))
    (cond
      ((atom? tail)
      ;((null? tail)
       (values (reverse head) tail))
      ((ok? (car tail))
       (values (reverse head) tail))
      (else
        (loop (cdr tail) (cons (car tail) head))))))

#|[
(take-while ok?)
(take-while ok? lst)
--- procedure ---
the second takes the first items of the list lst,
which pass the ok? test;
the first is a curried version of the second.
]|#
(define (take-while ok? . lists)
  (if (null? lists)
    (lambda (lst)
      (receive (head tail) (split-when (complement ok?) lst)
        head))
    (receive (head tail) (split-when (complement ok?) (car lists))
      head)))

#|[
(drop-while ok?)
(drop-while ok? lst)
--- procedure ---
the second drops the first items of the list lst,
which pass the ok? test;
the first is a curried version of the second.
]|#
(define (drop-while ok? . lists)
  (if (null? lists)
    (lambda (lst)
      (receive (head tail) (split-when (complement ok?) lst)
        tail))
    (receive (head tail) (split-when (complement ok?) (car lists))
      tail)))

#|[
(every? ok?)
(every? ok? xs)
--- procedure ---
checks, if every item of the list xs passes the ok? test.
The first is the curried version of the second call.
]|#
(define (every? ok? . rest)
  (cond
    ((null? rest)
     (lambda (xs) 
       (equal? (filter ok? xs) xs)))
    ((null? (cdr rest))
     (let ((xs (car rest)))
       (equal? (filter ok? xs) xs)))
    (else 
      (error 'every? "too many arguments" rest))))

#|[
(all? op?)
--- procedure ---
returns a predicate, which tests, if all items of it's only argument,
a list, are pairweise passed by the binary predicate op?)
]|#
(define (all? op?)
  (lambda (lst)
    (if (null? lst)
      #t
      ((list-of? (lambda (x) (op? x (car lst)))) (cdr lst)))))

#|[
(some? ok?)
(some? ok? xs)
--- procedure ---
checks, if some item of the list xs passes the ok? test.
The first is the curried version of the second call.
]|#
(define (some? ok? . rest)
  (cond
    ((null? rest)
     (lambda (xs) 
       (not (null? (filter ok? xs)))))
    ((null? (cdr rest))
     ((some? ok?) (car rest)))
    (else 
      (error 'some? "too many arguments" rest))))

#|[
(none? ok?)
(none? ok? xs)
--- procedure ---
checks, if no item of the list xs passes the ok? test.
The first is the curried version of the second call.
]|#
(define (none? ok? . rest)
  (cond
    ((null? rest)
     (lambda (xs) 
       (null? (filter ok? xs))))
    ((null? (cdr rest))
     (let ((xs (car rest)))
       (null? (filter ok? xs))))
    (else 
      (error 'none? "too many arguments" rest))))

#|[
(pseudo-list? xpr)
--- procedure ---
checks, if xpr evaluates to a pair, which is not a list,
or to an atom which is not null.
]|#
(define (pseudo-list? xpr)
  (or (and (pair? xpr) (not (list? xpr)))
      (and (atom? xpr) (not (null? xpr)))))

#|[
(pseudo-length pls)
--- procedure ---
returns the length of a pseudo-list or list
]|#
(define (pseudo-length pls)
  (let loop ((pls pls) (result 0))
    (if (atom? pls)
      result
      (loop (cdr pls) (+ 1 result)))))

#|[
(pseudo-head pls)
--- procedure ---
returns the head of a pseudo-list,
i.e. the list stripped-of the sentinel.
In case of a list, the list itself is returned.
]|#
(define (pseudo-head pls)
  (let loop ((pls pls))
    (if (pair? pls)
      (cons (car pls) (loop (cdr pls)))
      '())))

#|[
(pseudo-tail pls)
--- procedure ---
returns the sentinel of a pseudo-list,
i.e. null in the case of lists
]|#
(define (pseudo-tail pls)
  (let loop ((pls pls))
    (if (pair? pls)
      (loop (cdr pls))
      pls)))

#|[
(pseudo-list-of? ok?)
--- procedure ---
returns a procedure, which checks, if its only argument
is a pseudo-list, whose items pass the ok? test.
A pseudo-list is not a list, except ok? is null? or atom?
]|#
(define ((pseudo-list-of? ok?) xpr)
  (or (ok? xpr)
      ;(null? xpr)
      (and (pair? xpr)
           (ok? (car xpr))
           ((pseudo-list-of? ok?) (cdr xpr)))))

#|[
(indexes xs)
--- procedure ---
returns the indexes of the list xs.
]|#
(define (indexes xs)
  (let loop ((xs xs) (i 0) (result '()))
    (if (null? xs)
      (reverse result)
      (loop (cdr xs) (+ i 1) (cons i result)))))

;;; #|[
;;; (verbose-error-messages?)
;;; (verbose-error-messages? new)
;;; --- parameter ---
;;; defines or changes the verbosity of error-messages
;;; ]|#
;;; (define verbose-error-messages?
;;;   (make-parameter #f (lambda (x)
;;;                        (if (boolean? x)
;;;                          x
;;;                          #f))))

#|[
(tree-of? ok?)
--- procedure ---
returns a procedure, which checks, if its only argument
is a tree, whose items pass the ok? test.
]|#
(define ((tree-of? ok?) xpr)
  (or (ok? xpr)
      (and (pair? xpr)
           ((tree-of? ok?) (car xpr))
           ((tree-of? ok?) (cdr xpr)))))

#|[
(iterate fn k)
(iterate fn k arg)
--- procedure ---
The second applies the function fn successively k times to the argument
arg.
The first is a curried version of the second.
]|#
(define (iterate fn times . rest)
  (cond
    ((null? rest)
     (lambda (arg)
       (let loop ((k times))
         (if (zero? k)
           arg
           (fn (loop (- k 1)))))))
    ((null? (cdr rest))
     ((iterate fn times) (car rest)))
    (else (error 'iterate "too many arguments" rest))))
     
#|[
(map* fn tree . trees)
--- procedure ---
maps a function, fn, over a list of trees
]|#
(define (map* fn tree . trees)
  (if (null? trees)
    (let loop ((tree tree))
      (cond
        ((null? tree) tree)
        ((pair? tree)
         (if (pair? (car tree))
           (cons (loop (car tree))
                 (loop (cdr tree)))
           (cons (fn (car tree))
                 (loop (cdr tree)))))
        (else ;atom
         (fn tree))))
    (let loop ((trees (cons tree trees)))
      (cond
        ((every? null? trees)
         '())
        ((some? null? trees)
         (error 'map* "trees not similar" trees))
        (else
          (let ((cars (map car trees)) (cdrs (map cdr trees)))
            (cond
              ((every? pair? cars)
               (cons (loop cars) (loop cdrs)))
              ((every? atom? cars)
               (cons (apply fn cars) (loop cdrs)))
              ((every? null? cdrs)
               (apply fn cars))
              (else
                (error 'map* "trees not similar" trees)))))))))

#|[
(when-map* ok? fn tree)
--- procedure ---
map fn recursively over tree, but only on those expressions,
which pass the ok? test
]|#
(define (when-map* ok? fn tree)
  (let loop ((tree tree))
    (if (null? tree)
      tree
      (let ((x (car tree)) (xs (cdr tree)))
        (cond
          ((ok? x)
           (cons (fn x) (loop xs)))
          ((pair? x)
           (cons (loop x) (loop xs)))
          (else (cons x (loop xs))))))))

#|[
(map** fn pat . trees)
--- procedure ---
returns two values, mapping recursively pat and trees with fn.
]|#
(define (map** fn pat . trees)
  (values (map* fn pat) (apply map* fn trees)))

#|[
(flatten* pairs)
--- procedure ---
transforms a nested list of pairs into a flat
list of pairs
]|#
(define (flatten* pairs)
  (let loop ((pairs pairs))
    (if (none? pair? (map car pairs))
      pairs
      (loop
        (mappend (lambda (pair)
                   (if (pair? (car pair)) pair (list pair)))
                 pairs)))))

#|[
(zip2 xls)
--- procedure ---
transforms a list of pairs into a pair of lists
]|#
(define (zip2 xls)
  (list (map car xls) (map cadr xls)))

#|[
(unzip* trees)
--- procedure ---
transforms a list of equally shaped nested pseudo-lists into
a nested list of lists or atoms
]|#
(define (unzip* trees)
  (assert (or (null? trees)
              (null? (cdr trees))
              (every? (lambda (tree)
                        (same-shape? tree (car trees)))
                      (cdr trees))))
  (cond
    ((or (null? trees) (null? (car trees)))
     '())
    ((atom? (car trees))
     trees)
    ((pair? (car trees))
      (let ((tree (caar trees)))
        (cond
          ((atom? tree)
           (cons (map car trees) (unzip* (map cdr trees))))
          ((pair? tree) ;(list? tree)
           (cons (unzip* (map car trees))
                 (unzip* (map cdr trees))))
          )))))

;;; transforms a list of integers into a list of successive
;;; 1's and -1's
(define (ints->ones ints) ; internal
  (mappend (lambda (i)
             (cond
               ((zero? i)
                '())
               ((= i 1)
                (list i))
               ((= i -1)
                (list i))
               ((positive? i)
                (let loop ((k i) (result '()))
                  (if (zero? k)
                    result
                    (loop (- k 1) (cons 1 result)))))
               ((negative? i)
                (let loop ((k (- i)) (result '()))
                  (if (zero? k)
                    result
                    (loop (- k 1)
                          (cons -1 result)))))
               (else
                 (error 'ints->ones
                        "not a list of integers"
                        ints))
               ))
           ints))

#|[
(walk . ks)
--- procedure ---
creates a procedure, which accesses a tree along the
nonzero exact integers ks from right to left,
where 1 represents cdr, -1 car.
]|#
(define (walk . ks)
  (apply o (map (lambda (k)
                  (if (= k 1)
                    cdr
                    car))
                (ints->ones ks))))

#|[
(walk* . ks)
--- procedure ---
the same as walk, but the ks are accessed from left to right
]|#
(define (walk* . ks)
  (apply walk (reverse ks)))

#|[
(memp ok? lst)
--- procedure ---
returns the sublist of lst, whose first item passes the ok? test,
or #f if no item passes it.
]|#
(define (memp ok? lst)
  (let loop ((lst lst))
    (cond
      ((null? lst) #f)
      ((ok? (car lst)) lst)
      (else (loop (cdr lst))))))

#|[
(dups-remove =? lst)
--- procedure ---
removes duplicates from the list lst, compared by the binary function =?
]|#
(define (dups-remove =? lst)
  (let loop ((lst lst) (result '()))
    (cond
      ((null? lst) (reverse result))
      ((memp ;(cut =? (car lst) <>) (cdr lst))
         (lambda (x) (=? x (car lst))) (cdr lst))
       (loop (cdr lst) result))
      (else
        (loop (cdr lst) (cons (car lst) result))))))

#|[
(and? . xprs)
--- procedure ---
procedure version of and, so that apply is usable
]|#
(define (and? . xprs)
  (call-with-current-continuation
    (lambda (return)
      (let loop ((xprs xprs))
        (cond
          ((null? xprs) #t)
          ((is-true? (car xprs)) (loop (cdr xprs)))
          (else (return #f)))))))

(define (is-true? x) (if x #t #f))

#|[
(sublist lst from)
(sublist lst from upto)
--- procedure ---
returns the sublist of a list, lst, starting from indox
from included up to index upto excluded. 
If the argument upto is missing, the length of lst is used instead.
]|#
(define (sublist lst from . upto)
  (let ((upto (if (null? upto)
                (length lst)
                (car upto))))
    (let loop ((lst lst) (i 0) (result '()))
      (cond
        ((null? lst)
         (reverse result))
        ((= i upto)
         (reverse result))
        ((< i from)
         (loop (cdr lst) (+ i 1) result))
        (else
          (loop (cdr lst) (+ i 1) (cons (car lst) result)))))))

;;; #|[
;;; (pseudo->pathes pat)
;;; --- procedure ---
;;; transforms a flat pseudo-list, pat, into a flat list of pairs consisting
;;; of the pseudo-list's atoms and the pathes to reach these atoms.
;;; ]|#
(define (pseudo->pathes pat)
  (let recu ((pls pat) (lst '()))
    (if (pair? pls)
      (cons (list (car pls) (cons -1 lst))
            (recu (cdr pls) (cons 1 lst)))
      (list (list pls lst)))));(cons -1 lst))))))

#|[
(tree->pathes pat)
--- procedure ---
transforms a nested pseudo-list, pat, breadth-first into a flat list of
pairs consisting of the nested pseudo-list's items and the pathes to
reach these items
]|#
(define (tree->pathes pat)
  (receive (pairs atoms)
    (filter-pairs pair? (pseudo->pathes pat))
    (apply append
           atoms
           (map (lambda (p q)
                  (map (lambda (r)
                         (list (car r)
                               (append (cadr r) q)))
                       p))
                (map tree->pathes (map car pairs))
                (map cadr pairs)))))

#|[
(pathes item tree)
--- procedure ---
returns a list of pathes, where each path is a list of integers,
which, when fed to walk, would return the item in the tree.
]|#
(define (pathes item tree)
  (map cadr
    (filter-pairs (lambda (x) (equal? x item))
                  (tree->pathes tree))))

#|[
(same-shape? tree1 tree2)
--- procedure ---
checks, if both trees have the same shape, i.e. have items
in the same position.
]|#
(define (same-shape? tree1 tree2)
  (equal? (map cadr (tree->pathes tree1))
          (map cadr (tree->pathes tree2))))
#|[
(pseudo->accessors pat)
(pseudo->accessors pat walker)
--- procedure ---
transforms a flat pseudo-list, pat, into a flat list of triples
consisting of the pseudo-list's atoms, the procedures to reach these
atoms and the expressions describing these procedures.
If walker is not given, 'walk is assumed.
]|#
(define (pseudo->accessors pat . walker)
  (let ((walker (if (null? walker) 'walk (car walker))))
    (map (lambda (pair)
           `(,(car pair)
              ,(apply walk (cadr pair))
              (,walker ,@(cadr pair))))
         (pseudo->pathes pat))))

#|[
(tree->accessors pat)
(tree->accessors pat walker)
--- procedure ---
transforms a flat pseudo-list, pat, into a flat list of triples
consisting of the pseudo-list's atoms, the procedures to reach these
atoms and the expressions describing these procedures.
If walker is not given, 'walk is assumed.
]|#
(define (tree->accessors pat . walker)
  (let ((walker (if (null? walker) 'walk (car walker))))
    (map (lambda (pair)
           `(,(car pair)
              ,(apply walk (cadr pair))
              (,walker ,@(cadr pair))))
         (tree->pathes pat))))


#|[
(apply-accessors pat tree)
(apply-accessors pat tree map?)
--- procedure ---
applies or maps, if map? is true, all accessors to a tree.
]|#
(define (apply-accessors pat tree . map?)
  (let ((map? (if (null? map?) #f #t)))
    (map (lambda (proc)
           (if map? (map proc tree) (proc tree)))
         (map cadr (tree->accessors pat)))))

#|[
(all-match? tree)
((all-match? tree) xss)
--- procedure ---
Returns a procedure, which tests, if all accessors match its only
argument xss. Shows the item, the corresponding path and the
argument of the first failing accessor.
]|#
(define (all-match? tree)
  (lambda (xss)
    (let ((triples (tree->accessors tree)))
      (call-with-current-continuation
        (lambda (out)
          (apply and?
           (map (lambda (triple)
                  (if (condition-case ((cadr triple) xss)
                        ((exn)
                         (out #f)))
                    #t))
                triples)))))))

#|[
(tree-walkers)
(tree-walkers sym)
--- procedure ---
documentation procedure
]|#
(define tree-walkers
  (let (
    (alist '(
      (values->list
        macro:
        (values->list vals)
        "transforms values, vals, into a list"
        )
      (list->values
        procedure:
        (list->values lst)
        "transforms a list, lst, into a values"
        )
      (list-values
        procedure:
        (list-values proc . args)
        "transforms the resulting values of calling proc"
        "with args into a list."
        "Used in tests most of the time"
        )
      (always
        procedure:
        (always xpr)
        "returns a procedure of arbitrary many arguments,"
        "which always returns the value of xpr, whatever"
        "it's arguments are given"
        )
      (filter
        procedure:
        (filter ok? lst)
        "splits a list into two sublists, one that pass the ok? test and one that"
        "doesn't"
        )
      (filter*
        procedure:
        (filter* ok? tree)
        "filters a tree according to the predicate ok? respecting the tree"
        "structure"
        )
      (filter-pairs
        procedure:
        (filter-pairs ok? pairs)
        "splits a list of pairs into two sublists, one that pass the ok? test"
        "on the cars and one that doesn't"
        )
      (at
        procedure:
        (at n)
        (at n lst)
        "the second takes the nth  item of the list lst,"
        "the first is a curried version of the second."
        )
      (take
        procedure:
        (take n)
        (take n lst)
        "the second takes the first n items of the list lst,"
        "the first is a curried version of the second."
        )
      (drop
        procedure:
        (drop n)
        (drop n lst)
        "the second drops the first n items of the list lst,"
        "the first is a curried version of the second."
        )
      (split-at
        procedure:
        (split-at n pls)
        "splitting a pseudo-list in two sublists, head and tail, the sublist head"
        "excluding the nth item and the pseudolist tail starting with nth item."
        )
      (split-when
        procedure:
        (split-when ok? lst)
        "splitting a list in two sublists, head and tail, the sublist head before"
        "the first item passing the ok? test, the sublist tail starting at"
        "that item."
        )
      (take-while
        procedure:
        (take-while ok?)
        (take-while ok? lst)
        "the second takes the first items of the list lst,"
        "which pass the ok? test;"
        "the first is a curried version of the second."
        )
      (drop-while
        procedure:
        (drop-while ok?)
        (drop-while ok? lst)
        "the second drops the first items of the list lst,"
        "which pass the ok? test;"
        "the first is a curried version of the second."
        )
      (every?
        procedure:
        (every? ok?)
        (every? ok? xs)
        "checks, if every item of the list xs passes the ok? test."
        "The first is the curried version of the second call."
        )
      (all?
        procedure:
        (all? op?)
        "returns a predicate, which tests, if all items of it's only argument,"
        "a list, are pairweise passed by the binary predicate op?)"
        )
      (some?
        procedure:
        (some? ok?)
        (some? ok? xs)
        "checks, if some item of the list xs passes the ok? test."
        "The first is the curried version of the second call."
        )
      (none?
        procedure:
        (none? ok?)
        (none? ok? xs)
        "checks, if no item of the list xs passes the ok? test."
        "The first is the curried version of the second call."
        )
      (pseudo-list?
        procedure:
        (pseudo-list? xpr)
        "checks, if xpr evaluates to a pair, which is not a list,"
        "or to an atom which is not null."
        )
      (pseudo-length
        procedure:
        (pseudo-length pls)
        "returns the length of a pseudo-list or list"
        )
      (pseudo-head
        procedure:
        (pseudo-head pls)
        "returns the head of a pseudo-list,"
        "i.e. the list stripped-of the sentinel."
        "In case of a list, the list itself is returned."
        )
      (pseudo-tail
        procedure:
        (pseudo-tail pls)
        "returns the sentinel of a pseudo-list,"
        "i.e. null in the case of lists"
        )
      (pseudo-list-of?
        procedure:
        (pseudo-list-of? ok?)
        "returns a procedure, which checks, if its only argument"
        "is a pseudo-list, whose items pass the ok? test."
        "A pseudo-list is not a list, except ok? is null? or atom?"
        )
      (indexes
        procedure:
        (indexes xs)
        "returns the indexes of the list xs."
        )
      (tree-of?
        procedure:
        (tree-of? ok?)
        "returns a procedure, which checks, if its only argument"
        "is a tree, whose items pass the ok? test."
        )
      (iterate
        procedure:
        (iterate fn k)
        (iterate fn k arg)
        "The second applies the function fn successively k times to the argument"
        "arg."
        "The first is a curried version of the second."
        )
      (map*
        procedure:
        (map* fn tree . trees)
        "maps a function, fn, over a list of trees"
        )
      (when-map*
        procedure:
        (when-map* ok? fn tree)
        "map fn recursively over tree, but only on those expressions,"
        "which pass the ok? test"
        )
      (map**
        procedure:
        (map** fn pat . trees)
        "returns two values, mapping recursively pat and trees with fn."
        )
      (flatten*
        procedure:
        (flatten* pairs)
        "transforms a nested list of pairs into a flat"
        "list of pairs"
        )
      (zip2
        procedure:
        (zip2 xls)
        "transforms a list of pairs into a pair of lists"
        )
      (unzip*
        procedure:
        (unzip* trees)
        "transforms a list of equally shaped nested pseudo-lists into"
        "a nested list of lists or atoms"
        )
      (walk
        procedure:
        (walk . ks)
        "creates a procedure, which accesses a tree along the"
        "nonzero exact integers ks from right to left,"
        "where 1 represents cdr, -1 car."
        )
      (walk*
        procedure:
        (walk* . ks)
        "the same as walk, but the ks are accessed from left to right"
        )
      (memp
        procedure:
        (memp ok? lst)
        "returns the sublist of lst, whose first item passes the ok? test,"
        "or #f if no item passes it."
        )
      (dups-remove
        procedure:
        (dups-remove =? lst)
        "removes duplicates from the list lst, compared by the binary function =?"
        )
      (and?
        procedure:
        (and? . xprs)
        "procedure version of and, so that apply is usable"
        )
      (sublist
        procedure:
        (sublist lst from)
        (sublist lst from upto)
        "returns the sublist of a list, lst, starting from indox"
        "from included up to index upto excluded. "
        "If the argument upto is missing, the length of lst is used instead."
        )
      (tree->pathes
        procedure:
        (tree->pathes pat)
        "transforms a nested pseudo-list, pat, breadth-first into a flat list of"
        "pairs consisting of the nested pseudo-list's items and the pathes to"
        "reach these items"
        )
      (pathes
        procedure:
        (pathes item tree)
        "returns a list of pathes, where each path is a list of integers,"
        "which, when fed to walk, would return the item in the tree."
        )
      (same-shape?
        procedure:
        (same-shape? tree1 tree2)
        "checks, if both trees have the same shape, i.e. have items"
        "in the same position."
        )
      (pseudo->accessors
        procedure:
        (pseudo->accessors pat)
        (pseudo->accessors pat walker)
        "transforms a flat pseudo-list, pat, into a flat list of triples"
        "consisting of the pseudo-list's atoms, the procedures to reach these"
        "atoms and the expressions describing these procedures."
        "If walker is not given, 'walk is assumed."
        )
      (tree->accessors
        procedure:
        (tree->accessors pat)
        (tree->accessors pat walker)
        "transforms a flat pseudo-list, pat, into a flat list of triples"
        "consisting of the pseudo-list's atoms, the procedures to reach these"
        "atoms and the expressions describing these procedures."
        "If walker is not given, 'walk is assumed."
        )
      (apply-accessors
        procedure:
        (apply-accessors pat tree)
        (apply-accessors pat tree map?)
        "applies or maps, if map? is true, all accessors to a tree."
        )
      (all-match?
        procedure:
        (all-match? tree)
        ((all-match? tree) xss)
        "Returns a procedure, which tests, if all accessors match its only"
        "argument xss. Shows the item, the corresponding path and the"
        "argument of the first failing accessor."
        )
      (tree-walkers
        procedure:
        (tree-walkers)
        (tree-walkers sym)
        "with sym: documentation of exported symbol"
        "without sym: list of exported symbols"
        )
        ))
      )
      (case-lambda
        (() (map car alist))
        ((sym)
         (let ((pair (assq sym alist)))
           (if pair
             (for-each print (cdr pair))
             (print "Choose one of " (map car alist))))))))
)
